import os

import cv2
import numpy as np

from ocred.preprocessing import Preprocessor

path = "images/CosmosOne.jpg"


def test_scan():
    pre = Preprocessor(path)
    assert isinstance(pre.img, np.ndarray)

    img = pre.scan()
    assert isinstance(img, np.ndarray)

    img = pre.scan(save=True)
    assert isinstance(img, np.ndarray)
    assert os.path.exists("scanned.png")

    os.remove("scanned.png")

    img = cv2.imread(path)
    pre = Preprocessor(img)

    image = pre.scan()
    assert isinstance(image, np.ndarray)

    orig = img.copy()
    _ = pre.scan(inplace=True, overriden_image=img)
    assert not np.testing.assert_array_equal(img, orig)


def test_rotate():
    pre = Preprocessor(path)
    assert isinstance(pre.img, np.ndarray)

    img, median_angle = pre.rotate()
    assert isinstance(median_angle, float)
    assert isinstance(img, np.ndarray)

    img, median_angle = pre.rotate(save=True)
    assert isinstance(img, np.ndarray)
    assert isinstance(median_angle, float)
    assert os.path.exists("rotated.png")

    os.remove("rotated.png")

    img, median_angle = pre.rotate(overriden_image=img)
    img1, median_angle = pre.rotate(overriden_image=img)
    assert isinstance(median_angle, float)
    assert isinstance(img, np.ndarray)
    assert isinstance(img1, np.ndarray)
    np.testing.assert_array_equal(img, img1)

    orig = img.copy()
    _ = pre.rotate(inplace=True, overriden_image=img)
    assert not np.testing.assert_array_equal(img, orig)


def test_remove_noise():
    pre = Preprocessor(path)
    assert isinstance(pre.img, np.ndarray)

    img = pre.remove_noise()
    assert isinstance(img, np.ndarray)

    img = pre.remove_noise(save=True)
    assert isinstance(img, np.ndarray)
    assert os.path.exists("noise_free.png")

    os.remove("noise_free.png")

    img = pre.remove_noise(overriden_image=img)
    assert isinstance(img, np.ndarray)

    orig = img.copy()
    _ = pre.remove_noise(inplace=True, overriden_image=img)
    assert not np.testing.assert_array_equal(img, orig)


def test_thicken_font():
    pre = Preprocessor(path)
    assert isinstance(pre.img, np.ndarray)

    img = pre.thicken_font()
    assert isinstance(img, np.ndarray)

    img = pre.thicken_font(save=True)
    assert isinstance(img, np.ndarray)
    assert os.path.exists("thick_font.png")

    os.remove("thick_font.png")

    img = pre.thicken_font(overriden_image=img)
    assert isinstance(img, np.ndarray)

    orig = img.copy()
    _ = pre.thicken_font(inplace=True, overriden_image=img)
    assert not np.testing.assert_array_equal(img, orig)



import os

from ocred.ocr import OCR

path_scanned = "images/Page.png"
path_real = "images/CosmosOne.jpg"
path_sign_board = "images/signboard.jpg"
path_invoice = "images/1146-receipt.jpg"


def test_ocr_with_scanned_image():

    ocr = OCR(
        False,
        path_scanned,
    )

    assert ocr.path == path_scanned
    assert ocr.preprocess is False

    text = ocr.ocr_meaningful_text(save_output=True)

    assert isinstance(ocr.text, str)
    assert isinstance(text, str)
    assert text == ocr.text
    assert os.path.exists("OCR.png")
    assert os.path.exists("output.txt")
    assert not os.path.exists("preprocessed.png")
    assert not os.path.exists("audio.mp3")

    ocr.text_to_speech()

    assert os.path.exists("audio.mp3")

    os.remove("audio.mp3")
    os.remove("OCR.png")
    os.remove("output.txt")


def test_ocr_with_real_image():

    ocr = OCR(
        True,
        path_real,
    )

    assert ocr.path == "preprocessed.png"
    assert ocr.preprocess is True

    text = ocr.ocr_meaningful_text()

    assert isinstance(ocr.text, str)
    assert isinstance(text, str)
    assert text == ocr.text
    assert os.path.exists("OCR.png")
    assert os.path.exists("preprocessed.png")
    assert not os.path.exists("audio.mp3")

    ocr.text_to_speech()

    assert os.path.exists("audio.mp3")

    os.remove("audio.mp3")
    os.remove("OCR.png")
    os.remove("preprocessed.png")


def test_ocr_sign_board():
    ocr = OCR(
        False,
        path_sign_board,
    )

    assert ocr.path == path_sign_board
    assert ocr.preprocess is False

    text = ocr.ocr_sparse_text(save_output=True)

    assert isinstance(ocr.text, str)
    assert isinstance(text, str)
    assert text == ocr.text
    assert os.path.exists("OCR.png")
    assert os.path.exists("output.txt")
    assert not os.path.exists("preprocessed.png")
    assert not os.path.exists("audio.mp3")

    ocr.text_to_speech(lang="hi")

    assert os.path.exists("audio.mp3")

    os.remove("audio.mp3")
    os.remove("OCR.png")
    os.remove("output.txt")


def test_ocr_invoices():
    global path_invoice
    ocr = OCR(
        False,
        path_invoice,
    )

    assert ocr.path == path_invoice
    assert ocr.preprocess is False

    text = ocr.ocr_sparse_text()

    assert isinstance(ocr.text, str)
    assert isinstance(text, str)
    assert text == ocr.text
    assert os.path.exists("OCR.png")
    assert not os.path.exists("preprocessed.png")

    extracted_info = ocr.process_extracted_text_from_invoice()

    assert isinstance(extracted_info, dict)
    assert isinstance(ocr.extracted_info, dict)
    assert ocr.extracted_info == extracted_info
    assert (
        "price" in extracted_info
        and "date" in extracted_info
        and "place" in extracted_info
        and "order_number" in extracted_info
        and "phone_number" in extracted_info
        and "post_processed_word_list" in extracted_info
    ) is True
    assert isinstance(extracted_info["price"], str)
    assert isinstance(extracted_info["date"], list)
    assert len(extracted_info["date"]) == 0
    assert isinstance(extracted_info["place"], str)
    assert isinstance(extracted_info["phone_number"], list)
    assert len(extracted_info["phone_number"]) == 1
    assert isinstance(extracted_info["order_number"], int)
    assert isinstance(extracted_info["post_processed_word_list"], list)

    path_invoice = "images/1166-receipt.jpg"
    ocr = OCR(
        False,
        path_invoice,
    )

    assert ocr.path == path_invoice
    assert ocr.preprocess is False

    text = ocr.ocr_sparse_text()

    assert isinstance(ocr.text, str)
    assert isinstance(text, str)
    assert text == ocr.text
    assert os.path.exists("OCR.png")
    assert not os.path.exists("preprocessed.png")

    extracted_info = ocr.process_extracted_text_from_invoice()

    assert isinstance(extracted_info, dict)
    assert isinstance(ocr.extracted_info, dict)
    assert ocr.extracted_info == extracted_info
    assert (
        "price" in extracted_info
        and "date" in extracted_info
        and "place" in extracted_info
        and "order_number" in extracted_info
        and "phone_number" in extracted_info
        and "post_processed_word_list" in extracted_info
    ) is True
    assert isinstance(extracted_info["price"], str)
    assert isinstance(extracted_info["date"], list)
    assert len(extracted_info["date"]) == 1
    assert isinstance(extracted_info["place"], str)
    assert isinstance(extracted_info["phone_number"], list)
    assert len(extracted_info["phone_number"]) == 0
    assert isinstance(extracted_info["order_number"], str)
    assert isinstance(extracted_info["post_processed_word_list"], list)

    os.remove("OCR.png")



import math
from typing import Optional, Tuple, Union

import cv2
import numpy as np
import numpy.typing as npt
from scipy import ndimage
from skimage.filters import threshold_local


class Preprocessor:
    """
    Preprocesses an image and makes it ready for OCR.

    Args:
        image:
            Path of the image or a numpy array.

    Examples:
        >>> import cv2
        >>> from scipy import ndimage
        >>> from ocred import Preprocessor
        >>> # scan the image and copy the scanned image
        >>> preprocessed = Preprocessor("images/CosmosTwo.jpg")
        >>> # scan the image and copy the scanned image
        >>> scanned = preprocessed.scan(inplace=True)
        >>> orig = scanned.copy()
        >>> # remove noise
        ... noise_free = preprocessed.remove_noise(
        ...     inplace=True, overriden_image=scanned
        ... )
        >>> # thicken the ink to draw Hough lines better
        >>> thickened = preprocessed.thicken_font(
        ...     inplace=True, overriden_image=noise_free
        ... )
        >>> # calculate the median angle of all the Hough lines
        >>> _, median_angle = preprocessed.rotate(
        ...     inplace=True, overriden_image=thickened
        ... )
        >>> # rotate the original scanned image
        >>> rotated = ndimage.rotate(orig, median_angle)
        >>> # remove noise again
        >>> final_img = preprocessed.remove_noise(inplace=True, overriden_image=rotated)
        >>> cv2.imwrite("preprocessed.png", final_img)
        True
    """

    def __init__(
        self,
        image: Union[str, Union[npt.NDArray[np.int64], npt.NDArray[np.float64]]],
    ) -> None:
        if isinstance(image, str):
            self.img = cv2.imread(image)
        else:
            self.img = image

    def remove_noise(
        self,
        *,
        save: Optional[bool] = False,
        inplace: Optional[bool] = False,
        iterations: Optional[int] = 1,
        overriden_image: Union[
            npt.NDArray[np.int64], npt.NDArray[np.float64], None
        ] = None
    ) -> Union[npt.NDArray[np.int64], npt.NDArray[np.float64]]:
        """
        Removes noise from an image.

        Args:
            save:
                Saves the resultant image.
            inplace:
                Edits the image inplace.
            iterations:
                Number of times the image is processed.
            overriden_image:
                Overrides the image passes through the constructor.

        Returns:
            noise_free_image:
                The noise free image.
        """
        if not inplace:
            img = self.img.copy() if overriden_image is None else overriden_image.copy()
        else:
            img = self.img if overriden_image is None else overriden_image

        kernel: Union[npt.NDArray[np.int64]] = np.ones((1, 1), np.uint8)
        img = cv2.dilate(img, kernel, iterations=iterations)
        kernel = np.ones((1, 1), np.uint8)
        img = cv2.erode(img, kernel, iterations=iterations)
        img = cv2.morphologyEx(img, cv2.MORPH_CLOSE, kernel)
        img = cv2.medianBlur(img, 3)

        if save:
            cv2.imwrite("noise_free.png", img)

        return self.img

    def thicken_font(
        self,
        *,
        save: Optional[bool] = False,
        inplace: Optional[bool] = False,
        iterations: Optional[int] = 2,
        overriden_image: Union[
            npt.NDArray[np.int64], npt.NDArray[np.float64], None
        ] = None
    ) -> Union[npt.NDArray[np.int64], npt.NDArray[np.float64]]:
        """
        Thickens the ink of an image.

        Args:
            save:
                Saves the resultant image.
            inplace:
                Edits the image inplace.
            iterations:
                Number of times the image is processed.
            overriden_image:
                Overrides the image passes through the constructor.

        Returns:
            thickened_image:
                The thickened image.
        """
        if not inplace:
            img = self.img.copy() if overriden_image is None else overriden_image.copy()
        else:
            img = self.img if overriden_image is None else overriden_image

        img = cv2.bitwise_not(img)
        kernel: Union[npt.NDArray[np.int64]] = np.ones((2, 2), np.uint8)
        img = cv2.dilate(img, kernel, iterations=iterations)
        img = cv2.bitwise_not(img)

        if save:
            cv2.imwrite("thick_font.png", img)

        return img

    def scan(
        self,
        *,
        save: Optional[bool] = False,
        inplace: Optional[bool] = False,
        overriden_image: Union[
            npt.NDArray[np.int64], npt.NDArray[np.float64], None
        ] = None
    ) -> Union[npt.NDArray[np.int64], npt.NDArray[np.float64]]:
        """
        Transforms an image/document view into B&W view (proper scanned colour scheme).

        Args:
            save:
                Saves the resultant image.
            inplace:
                Edits the image inplace.
            overriden_image:
                Overrides the image passes through the constructor.

        Returns:
            scanned_image:
                The scanned image.
        """
        if not inplace:
            img = self.img.copy() if overriden_image is None else overriden_image.copy()
        else:
            img = self.img if overriden_image is None else overriden_image

        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        thr = threshold_local(img, 11, offset=10, method="gaussian")
        img = (img > thr).astype("uint8") * 255

        if save:
            cv2.imwrite("scanned.png", img)

        return img

    def rotate(
        self,
        *,
        save: Optional[bool] = False,
        inplace: Optional[bool] = False,
        overriden_image: Union[
            npt.NDArray[np.int64], npt.NDArray[np.float64], None
        ] = None
    ) -> Tuple[Union[npt.NDArray[np.int64], npt.NDArray[np.float64]], float]:
        """
        Rotates an image for a face-on view (view from the top).

        Args:
            save:
                Saves the resultant image.
            inplace:
                Edits the image inplace.
            overriden_image:
                Overrides the image passes through the constructor.

        Returns:
            rotated_image:
                The rotated image.
            median_angle:
                The angly by which it is rotated.
        """
        if not inplace:
            img = self.img.copy() if overriden_image is None else overriden_image.copy()
        else:
            img = self.img if overriden_image is None else overriden_image

        img_edges = cv2.Canny(img, 100, 100, apertureSize=3)
        lines = cv2.HoughLinesP(
            img_edges,
            rho=1,
            theta=np.pi / 180.0,
            threshold=160,
            minLineLength=100,
            maxLineGap=10,
        )

        angles = []
        for [[x1, y1, x2, y2]] in lines:
            angle = math.degrees(math.atan2(y2 - y1, x2 - x1))
            angles.append(angle)

        median_angle = float(np.median(angles))
        img = ndimage.rotate(img, median_angle)

        if save:
            cv2.imwrite("rotated.png", img)

        return img, median_angle



import nox

ALL_PYTHONS = ["3.7", "3.8", "3.9"]

nox.options.sessions = ["lint", "tests", "doctests"]


@nox.session(reuse_venv=True)
def lint(session):
    """Run the linter."""
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session(python=ALL_PYTHONS, reuse_venv=True)
def tests(session):
    """Run the unit and regular tests."""
    session.install(".[dev]")
    session.run("pytest", *session.posargs)


@nox.session(reuse_venv=True)
def doctests(session):
    """Run the doctests."""
    session.install(".[dev]")
    session.run("xdoctest", "./ocred/", *session.posargs)


@nox.session(reuse_venv=True)
def docs(session):
    """Build the docs. Pass "serve" to serve."""
    session.install("-e", ".[docs]")

    if session.posargs:
        if "serve" in session.posargs:
            print("Launching docs at http://localhost:8000/ - use Ctrl-C to quit")
            session.run("mkdocs", "serve")
        else:
            print("Unsupported argument to docs")
    else:
        session.run("mkdocs", "build")


@nox.session
def build(session):
    """Build an SDist and wheel."""
    session.install("build")
    session.run("python", "-m", "build")



import re
import warnings
from typing import Any, Dict, List, Optional, Union

import cv2
import easyocr
import pytesseract
from gtts import gTTS
from scipy import ndimage

from ocred.preprocessing import Preprocessor


class OCR:
    """
    Performs OCR on a given image, saves an image with boxes around the words, and
    converts the extracted text to an MP3 file.

    Add Tesseract OCR's installation location in PATH for functions using it to work.

    Args:

        preprocess:
            Set True if the image is a real life photo of some large meaningful (page of a
            book). Usually set to False when OCRing using `ocr_meaningful_text` to
            preprocess the image.
            Set False if the image is a scanned photo (an e-book). It will not be
            pre-processed before OCRing.
            All the preprocessing is performed inplace to maintain efficiency. Use the
            `Preprocessor` class manually to have more control!
        path:
            Path of the image to be used.

    Examples:

        >>> import sys
        >>> sys.displayhook = lambda x: None
        >>> import ocred
        >>> ocr = ocred.OCR(
        ...     False, # is_scanned -> to preprocess the image
        ...     "./images/Page.png"
        ... )
        >>> ocr.ocr_meaningful_text(save_output=True)
    """

    def __init__(self, preprocess: bool, path: str) -> None:
        self.path = path
        self.preprocess = preprocess

        if self.preprocess:
            preprocessed = Preprocessor(self.path)

            # scan the image and copy the scanned image
            scanned = preprocessed.scan(inplace=True)
            orig = scanned.copy()

            # remove noise
            noise_free = preprocessed.remove_noise(
                inplace=True, overriden_image=scanned
            )

            # thicken the ink to draw Hough lines better
            thickened = preprocessed.thicken_font(
                inplace=True, overriden_image=noise_free
            )

            # calculate the median angle of all the Hough lines
            _, median_angle = preprocessed.rotate(
                inplace=True, overriden_image=thickened
            )

            # rotate the original scanned image
            rotated = ndimage.rotate(orig, median_angle)

            # remove noise again
            final_img = preprocessed.remove_noise(inplace=True, overriden_image=rotated)

            cv2.imwrite("preprocessed.png", final_img)
            self.path = "preprocessed.png"

    def ocr_meaningful_text(self, *, save_output: Optional[bool] = False) -> str:
        """
        Performs OCR on long meaningful text documents and saves the image with boxes
        around the words. For example - books, PDFs etc.

        Args:
            save_output:
                Saves the text to `output.txt` file.

        Returns:
            text:
                The extracted text.
        """
        # reading the image
        img = cv2.imread(self.path)

        # extracting the text
        self.oriented_text = pytesseract.image_to_string(img, config="-l eng --oem 1")
        self.text = self.oriented_text.replace("-\n", "").replace("\n", " ")

        # adding boxes around the words
        boxes = pytesseract.image_to_data(img)
        for z, box in enumerate(boxes.splitlines()):
            if z != 0:
                box = box.split()

                # if the data has a word
                if len(box) == 12:

                    x, y = int(box[6]), int(box[7])
                    h, w = int(box[8]), int(box[9])

                    cv2.rectangle(img, (x, y), (x + h, y + w), (0, 0, 255), 1)

        cv2.imwrite("OCR.png", img)

        if save_output:
            self.save_output()

        return self.text

    def ocr_sparse_text(
        self,
        *,
        languages: Optional[List[str]] = ["en", "hi"],
        decoder: Optional[str] = "greedy",
        save_output: Optional[bool] = False,
    ) -> str:
        """
        Performs OCR on sparse text and saves the image with boxes around the words.
        This method can be used to OCR documents in which the characters don't form any
        proper/meaningful sentences, or if there are very less meaningful sentences,
        for example - bills, sign-boards etc.

        Args:
            languages:
                A list of languages that the signboard possible has.
                Note: Provide only the languages that are present in the image, adding
                additional languages misguides the model.
            decoder:
                If the document has a larger number of meaningful sentences then use
                "beamsearch". For most of the cases "greedy" works very well.
            save_output:
                Saves the text to `output.txt` file.

        Returns:
            text:
                The extracted text.
        """
        self.text = ""

        # reading the image using open-cv and easyocr
        img = cv2.imread(self.path)
        reader = easyocr.Reader(
            languages
        )  # slow for the first time (also depends upon CPU/GPU)
        self.detailed_text = reader.readtext(self.path, decoder=decoder, batch_size=5)

        for text in self.detailed_text:

            # extracting the coordinates to highlight the text
            coords_lower = text[0][:2]
            coords_upper = text[0][2:4]

            coords_lower.sort(key=lambda x: x[0])
            pt1 = [int(x) for x in coords_upper[-1]]

            coords_lower.sort(key=lambda x: x[0])
            pt2 = [int(x) for x in coords_lower[-1]]

            # highlighting the text
            cv2.rectangle(img, pt1, pt2, (0, 0, 255), 1)

            self.text = self.text + " " + text[-2]

        cv2.imwrite("OCR.png", img)

        if save_output:
            self.save_output()

        return self.text

    def process_extracted_text_from_invoice(self) -> Dict[str, Any]:
        """
        This method processes the extracted text from invoices, and returns some useful
        information.

        Returns:
            extracted_info:
                The extracted information.
        """
        import nltk

        nltk.download("punkt")
        nltk.download("wordnet")
        nltk.download("stopwords")

        self.extracted_info = {}
        self.text_list = self.text.split(" ")

        # find date
        date_re = re.compile(
            r"^([1-9]|0[1-9]|1[0-9]|2[0-9]|3[0-1])(\.|-|\/)([1-9]|0[1-9]|1[0-2])(\.|-|\/)([0-9][0-9]|19[0-9][0-9]|20[0-9][0-9])$",
        )
        date = list(filter(date_re.match, self.text_list))

        # find phone number
        phone_number_re = re.compile(
            r"((\+*)((0[ -]*)*|((91 )*))((\d{12})+|(\d{10})+))|\d{5}([- ]*)\d{6}",
        )
        phone_number = list(filter(phone_number_re.match, self.text_list))

        # find place
        place = self.detailed_text[0][-2]

        # remove puntuations and redundant words
        tokenizer = nltk.RegexpTokenizer(r"\w+")
        removed_punctuation = tokenizer.tokenize(self.text)

        stop_words = set(nltk.corpus.stopwords.words("english"))
        post_processed_word_list = [
            w for w in removed_punctuation if w not in stop_words
        ]

        # find order number
        order_number: Union[str, int] = ""
        for i in range(len(post_processed_word_list)):
            if post_processed_word_list[i].lower() == "order":
                try:
                    order_number = int(post_processed_word_list[i + 1])
                except Exception:
                    order_number = post_processed_word_list[i + 2]
                break

        # find total price
        price: Union[List[Any], str] = ""

        # try finding a number with Rs, INR, ₹ or रे in front of it or Rs, INR at the end
        # of it
        try:
            price = re.findall(
                r"(?:Rs\.?|INR|₹\.?|रे\.?)\s*(\d+(?:[.,]\d+)*)|(\d+(?:[.,]\d+)*)\s*(?:Rs\.?|INR)",
                self.text,
            )
            price = list(map(float, price))
            price = max(price)
        # try finding numbers with "grand total" or "total" written in front of them
        except ValueError:
            lowered_list = [x.lower() for x in post_processed_word_list]
            if "grand" in lowered_list:
                indices = [i for i, x in enumerate(lowered_list) if x == "grand"]
                i = indices[-1]
                price = post_processed_word_list[i + 2]
            elif "total" in lowered_list:
                indices = [i for i, x in enumerate(lowered_list) if x == "total"]
                i = indices[-1]
                price = post_processed_word_list[i + 1]

        self.extracted_info.update(
            {
                "price": price,
                "date": date,
                "place": place,
                "order_number": order_number,
                "phone_number": phone_number,
                "post_processed_word_list": post_processed_word_list,
            }
        )

        return self.extracted_info

    def save_output(self) -> None:
        """Saves the extracted text in the `output.txt` file."""
        f = open("output.txt", "w", encoding="utf-8")
        f.write(self.text)
        f.close()

    def text_to_speech(self, *, lang: Optional[str] = "en") -> None:
        """
        Converts the extracted text to speech and save it as an MP3 file.

        Args:
            lang:
                Language of the processed text.
        """
        warnings.simplefilter("always", DeprecationWarning)
        warnings.warn(
            "text_to_speech is deprecated and will be removed in v0.2.0; use gTTS manually",
            DeprecationWarning,
        )

        speech = gTTS(self.text, lang="en", tld="com")
        speech.save("audio.mp3")
